    org &8000
    run &8000

; kill system interrupt handler first, the alternate register set is used 
    di
    ld hl,&c9fb
    ld (&0038),hl
    ei

; set mode 0
    ld bc,&7F8c
    out (c),c

; set colours
    ld a,15
    ld hl,Mode0Pal
    call SetColours

; clear both screens
    ld a,&c0
    call ClearScr
    ld a,&40
    call ClearScr

; set screen size & position
    ld hl,Mode0Dim
    call SetScreen

; ensure screen starts high (&c000)
    call ShowHigh

.MainLoop
    call SwapScr

    halt
    halt

    call ProcessBGList

; throttle to 25hz
    call FrameFlyB

    halt
    halt

; uncomment out these 8 lines to see how long this code takes
;ld bc,&7F10
;out (c),c
;ld c,&55
;out (c),c
    call PrintBGList
;ld bc,&7F10
;out (c),c
;ld c,&54
;out (c),c

    jr MainLoop

list
.SwapScr
nolist
    call FrameFlyB
; set R3
    ld a,(R3Setting)
    ld bc,&bc03
    out (c),c
    inc b
    out (c),a
    ld d,a
; set screen base
    ld a,(WorkScr)
; store screen base address in e and modify a for hardware output
    ld e,a
    srl a
    srl a
    ld bc,&bc0c
    out (c),c
    inc b
    out (c),a
; retrieve screen base address from e, and swap to new work screen now that
; old work screen is being displayed
    ld a,e
    xor &80 ; alt between &c0,&40
    ld (WorkScr),a
; set course offset
    ld hl,ScrollOffset
    ld bc,&bc0d
    ld a,(hl)
    out (c),c
    inc b
    out (c),a
; put offset into c for later checks
    ld c,a
; scroll code is from star sabre, so it checks to see
; if we need to stop scrolling - not used in this example
    ld a,(ScrollOn)
    or a
    jr z,StopScrollCheck
.KeepScrolling
; if current screen is low screen, we return as no further hardware settings
; need changing, scrolling this frame is achieved with preshifted blocks
    bit 7,e
    ret z
; change Register 3
    ld a,3
    xor a,d
    ld (R3Setting),a
    ld a,(R3Setting)
    cp a,&86
    ret z
; if just moved screen half char right (&85), inc course offset so actual shift
; is half char left
    ld a,c
    inc a
    cp a,181
    jp nz,SwScSkip
; if reached maximum offset, reset for next background list
; this means each set of screens needs to be seperated by blank screen
; but avoids end of screen memory address checks for all screen writes
    ld a,48
    ld (BlockFraction),a
    xor a
    ld (BlockPause),a
.SwScSkip
; write new offset and return
    ld (hl),a
    ret
.StopScrollCheck
; if scrolling is to be turned off, stop scrolling when screen offset
; is next reset to 0
    xor a
    cp a,c
    ret z
    jr KeepScrolling

list
.ProcessBGList
nolist
; dont process list if at low screen (&4000) or scrolling turned off 
    ld a,(WorkScr)
    rla
    ret nc
    ld a,(ScrollOn)
    or a
    ret z
; increment and check block fraction (0-3 byte columns of 16 bytes)
    ld hl,BlockFraction
    ld a,(hl)
    add a,16
    and a,&3f
    ld (hl),a
    ret nz
; if block fraction resets to zero, check pause state
    ld hl,BlockPause
    ld a,(hl)
    or a
    jr z,ProBGLNoPause
; if pausing, decrement blockpause and return
; pausing is used to hold the list processing between background segements
; as well as allowing for repeated columns
    dec a
    ld (hl),a
    ret nz
.ProBGLNoPause
; add length of previous block list to pointer
    ld hl,(BlockPointer)
    ld a,(BlockCount)
    inc a ; Previous number of blocks +1 for instruction byte
    ld d,0
    ld e,a
    add hl,de
.ProBGListStart
    ld (BlockPointer),hl
; read first byte in block list. This has high bit set for special instruction
; or reset to define a list of blocks, defined in bits as 0cccnnnn
; where ccc = number of times to print column (1-7)
; nnnn = number of blocks in list
    ld a,(hl)
    ld e,a
    bit 7,a
    jr nz,ProBGLSpecialCase
    and a,&70
    rrca
    rrca
    rrca
    rrca
; write column count to blockpause, causing list processor to wait ccc block widths
; must not equal zero (will try to repeat column 255 times)
    ld (BlockPause),a
    ld a,e
    and a,15
; and block count to variable so print routine knows how many of next bytes to
; process, and how many bytes to add when next column is read
    ld (BlockCount),a
    ret
.ProBGLSpecialCase
; if special case, only low 4 bits matter
    and a,15
    cp a,13
    jr z,ProLoopBG
    cp a,14
    jr z,ProWaitBG
; assume 15 at this point
; write 255 to block pause so list processor will be wait until
; screen course offset is reset
    ld a,255
.SkipLongLp
    ld (BlockPause),a
.ProBGListEnd
; signify no blocks to print
    xor a
    ld (BlockCount),a
    ret
.ProWaitBG
; flag to turn scrolling off in event of boss encounter
    xor a
    ld (ScrollOn),a
    jr ProBGListEnd
.ProLoopBG
; read next two bytes and subtract from block list pointer so can loop background
    inc hl
    ld e,(hl)
    inc hl
    ld d,(hl)
    sbc hl,de
    jr ProBGListStart

.PrintBGList
; first load de' with start address for column to print
    exx
; 'scrolloffset' is course offset in one byte, with base address stored in
; 'workscr', which is &40 or &c0.  To get working address, register l is doubled
; with a left shift and inc h if it carries, then inc hl depending on R3 setting
    ld hl,(ScrollOffset)
    ld c,l
    sla l
    jr nc,PrnBGLSkip1
    inc h
.PrnBGLSkip1
; 'nudge' the column to be written right by one byte if R3 = &86 so that some
; for those frames the graphics write to second last column of bytes on
; screen and blank the last column of bytes, and on frames where R3 = &85 the
; graphics are written to last column of bytes, and the first column of bytes
; on screen is blanked - except for top character line, which is dealt with shortly
    ld a,(R3Setting)
    rra
    jr c,PrnBGLR3SkipNudge
    inc hl
.PrnBGLR3SkipNudge
; save hl in case it is needed to blank top left character line
    push hl
; and put offset in a so we skip if zero
    ld a,c
    ld de,78
    add hl,de
    ex de,hl
; de' now contains column of bytes to write to
    exx
    pop hl
    or a
    jr z,DontBlankLeftChar
; if scrolloffset is not zero, clear 8 bytes in column on top left of screen
; so when offset is eventually reset, screen doesnt fill with junk
    dec hl
    xor a
    ld (hl),a
    set 3,h
    ld (hl),a
    set 4,h
    ld (hl),a
    res 3,h
    ld (hl),a
    set 5,h
    ld (hl),a
    set 3,h
    ld (hl),a
    res 4,h
    ld (hl),a
    res 3,h
    ld (hl),a
.DontBlankLeftChar
; if block count is zero, just write a blank column to de'
    ld a,(BlockCount)
    or a
    jp z,BlankColumn
; if we use the blank blocks routine from this point, it needs to
; know to return to the block loop
    ld ix,PrnBlocksEnd
    ld e,a ; put block count in e
; work out which set of preshifted blocks to use based on current work screen
; base address and set the blocks base in bc
    ld a,(WorkScr)
    rlca
    jr c,PrnBGLHighScrn
    ld bc,LowBlocks
    jr PrnBGLSkipLow
.PrnBGLHighScrn
    ld bc,HighBlocks
.PrnBGLSkipLow
    ld a,e ; get block count back from e
    ld de,(BlockPointer)
    inc de ; set de to point to first block in list
.PrnBGOuterLoop
    ex af,af' ; save block list count for end of loop, so use a'
    ld a,(de) ; get low 6 bits for block number
    and a,&3f ; block zero is blank, use seperate routine
    jr z,PrnZeroBlocks
    dec a ; reduce block number to range 0-62, rather than 1-63
; multiply a by 64 and put in hl
; this is done by putting a/4 in h, with the 2 ls bits going into the top of l
    rrca
    rrca
    ld l,a
    and a,&f
    ld h,a
    ld a,l
    and a,&c0
    ld l,a
; add blockfraction (0-48) to hl
    ld a,(BlockFraction)
    add a,l
    ld l,a
; add block graphics base previously put in bc to hl
    add hl,bc
; and push onto stack so can be retrieved by hl'
    push hl
; now get top 2 bits, this determines how many times to print the block
    ld a,(de)
    rlca
    rlca
    and a,3
    inc a ; print 1-4 times, not 0-3 times
    exx
; retrieve block graphics to write
    pop hl
    add a,a ; each block is two character lines, so find character lines to print
; put block count into b to use as counter and zero a
    ld b,a
    xor a
    ld c,255 ; prevent ldi from corrupting b
.PrnBlockLoop
; block data is stored in such a way that we only need one set or reset to
; get to each new line, so on screen you see a block printed:
; row 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15
; but data is stored:
; row 0,1,3,2,6,7,5,4,8,9,11,10,14,15,13,12
    ldi
    ld (de),a
    set 3,d
    dec de
    ldi
    ld (de),a
    set 4,d
    dec de
    ldi
    ld (de),a
    res 3,d
    dec de
    ldi
    ld (de),a
    set 5,d
    dec de
    ldi
    ld (de),a
    set 3,d
    dec de
    ldi
    ld (de),a
    res 4,d
    dec de
    ldi
    ld (de),a
    res 3,d
    dec de
    ldi
    ld (de),a
; back to top scan line of character
    res 5,d
; beginning of next character line is 80 (-1 for last ldi) bytes away
    ld a,79
    add a,e
    jr nc,PrnBlkDontIncD
    inc d
.PrnBlkDontIncD
    ld e,a
    xor a
; new line
    dec b
; if that was last character line, b=0 and finished with printing this block
    jr z,PrnBlocksEnd
; if still going, check to see if its the start of a new repeat block
    bit 0,b
    jr nz,PrnBlockLoop ; if b is odd number, halfway through block
; b is even, must be repeating same block now, go back to hl and push/pop it again
    exx
    push hl
    exx
    pop hl
; check for block repeat
    jr PrnBlockLoop
.PrnBlocksEnd ; return from below blank block routine here
    exx
; finished printing that block(s) so move de to next block
    inc de
; get back to a and see if we have more blocks in column list to print
    ex af,af'
    dec a
    jr nz,PrnBGOuterLoop
    ret
; now for blank block
.PrnZeroBlocks
; this process is the same as printing a block essentially, except we have no
; source block
;
; still read block count 1-4 for blank blocks
    ld a,(de)
    rlca
    rlca
    and a,3
    inc a
.BlankColumnEntry
    exx
    ld c,a
    xor a
    ld hl,&50
; swap destination into hl for blanks
    ex de,hl
.PrnBlankLoopOuter
    ld b,2
.PrnBlankLoopInner
    ld (hl),d
    inc hl
    ld (hl),d
    set 3,h
    ld (hl),d
    dec hl
    ld (hl),d
    set 4,h
    ld (hl),d
    inc hl
    ld (hl),d
    res 3,h
    ld (hl),d
    dec hl
    ld (hl),d
    set 5,h
    ld (hl),d
    inc hl
    ld (hl),d
    set 3,h
    ld (hl),d
    dec hl
    ld (hl),d
    res 4,h
    ld (hl),d
    inc hl
    ld (hl),d
    res 3,h
    ld (hl),d
    dec hl
    ld (hl),d
; back to top scan line of character
    res 5,h
    add hl,de
; new line
    djnz PrnBlankLoopInner
    dec c
    jr nz,PrnBlankLoopOuter
; swap destination back into de when blanks done
    ex de,hl
; return to calling function
    jp (ix)

list
.BlankColumn
nolist
; if we use the blank blocks routine from this point, it needs to
; know to return here, not the main block loop
    ld ix,BlankColumnReturn
; now we just set up the details for getting the blank block routine to print
; 10 blank blocks
    ld a,10
    jr BlankColumnEntry
.BlankColumnReturn
    exx
    ret

list
.FrameFlyB
nolist
    ld b,&f5
.FFBLoop
    in a,(c)
    rra
    jr nc,FFBLoop
    ret

list
.ShowHigh
nolist
    ld a,&30
    jr DontShowLow
list
.ShowLow
nolist
    ld a,&10
.DontShowLow
    ld bc,&bc0c
    out (c),c
    inc b
    out (c),a
    dec b
    inc c
    xor a
    out (c),c
    inc b
    out (c),a
    ret

list
.SetScreen
nolist
; HL points to 8 byte list for OUT
    ld d,8
    ld c,0
    ld a,&bc
.SetScrLoop
    ld e,(hl)
    ld b,a
    out (c),e
    inc hl
    xor 1 ; alts &bc to &bd in A
    dec d
    jr nz,SetScrLoop
    ret

list
.SetColours
nolist
; HL points to list, A holds 15 or 3

    ld b,&7f
    ld c,(hl)
    out (c),a
    out (c),c
    inc hl
    or a
    jr z,SetBorder
    dec a
    jr SetColours
.SetBorder
    ld a,&10
    out (c),a
    out (c),c
    ret

list
.ClearScr
nolist
; A holds &c0 or &40
    ld h,a
    ld l,0
    ld b,128
    xor a
.CSOuterLp
    ld c,128
.CSInnerLp
    ld (hl),a
    inc hl
    dec c
    jr nz,CSInnerLp
    dec b
    jr nz,CSOuterLp
    ret

.Mode0Pal
    defb &58,&4d,&4f,&4c,&4e,&4a,&56,&52,&54,&55,&57,&53,&44,&5c,&4b,&54

.Mode0Dim
    defb 2,46,1,40,7,28,6,21

.TestLevel
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+7,0+1,0+2,0+1,192+0,0+2,0+1,0+2
    defb 16+7,0+2,0+1,0+2,192+0,0+1,0+2,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+7,0+1,0+2,0+1,192+0,0+2,0+1,0+2
    defb 16+7,0+2,0+1,0+2,192+0,0+1,0+2,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+7,0+1,0+2,0+1,192+0,0+2,0+1,0+2
    defb 16+7,0+2,0+1,0+2,192+0,0+1,0+2,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+7,0+1,0+2,0+1,192+0,0+2,0+1,0+2
    defb 16+7,0+2,0+1,0+2,192+0,0+1,0+2,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+7,0+1,0+2,0+1,192+0,0+2,0+1,0+2
    defb 16+7,0+2,0+1,0+2,192+0,0+1,0+2,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+7,0+1,0+2,0+1,192+0,0+2,0+1,0+2
    defb 16+7,0+2,0+1,0+2,192+0,0+1,0+2,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+4,0+1,192+0,192+0,0+2
    defb 16+4,0+2,192+0,192+0,0+1
    defb 16+7,0+1,0+2,0+1,192+0,0+2,0+1,0+2
    defb 16+7,0+2,0+1,0+2,192+0,0+1,0+2,0+1
    defb 143 ; wait until offset is zero

; new level segment using repeated columns
    defb 64+4,0+1,192+0,192+0,0+2
    defb 64+4,0+2,192+0,192+0,0+1
    defb 32+7,0+1,0+2,0+1,192+0,0+1,0+2,0+1
    defb 64+4,0+1,192+0,192+0,0+2
    defb 64+4,0+2,192+0,192+0,0+1
    defb 32+7,0+1,0+2,0+1,192+0,0+1,0+2,0+1
    defb 64+4,0+1,192+0,192+0,0+2
    defb 64+4,0+2,192+0,192+0,0+1
    defb 32+7,0+1,0+2,0+1,192+0,0+1,0+2,0+1
    defb 64+4,0+1,192+0,192+0,0+2
    defb 64+4,0+2,192+0,192+0,0+1
    defb 32+7,0+1,0+2,0+1,192+0,0+1,0+2,0+1
    defb 64+4,0+1,192+0,192+0,0+2
    defb 64+4,0+2,192+0,192+0,0+1
    defb 32+7,0+1,0+2,0+1,192+0,0+1,0+2,0+1
    defb 64+4,0+1,192+0,192+0,0+2
    defb 64+4,0+2,192+0,192+0,0+1
    defb 32+7,0+1,0+2,0+1,192+0,0+1,0+2,0+1
    defb 64+4,0+1,192+0,192+0,0+2
    defb 64+4,0+2,192+0,192+0,0+1
    defb 32+7,0+1,0+2,0+1,192+0,0+1,0+2,0+1
    defb 143 ; wait until offset is zero
    defb 141 ; loop back to beginning
    defw EndTestLevel-TestLevel-1
.EndTestLevel

.ScrollOffset ; screen course offset
    defb 0
.WorkScr ; screen base address, either &40 or &c0
    defb &40
.R3Setting ; R3 setting = &85 or &86 
    defb &86
.ScrollOn ; allows scroll to be halted when offset = 0
    defb 1
.BlockFraction ; fraction of block being printed = 0,16,32,48
    defb 48
.BlockCount ; number of blocks in current column
    defb 0
.BlockPause ; number of times to print current column (including blank columns) 
    defb 0
.BlockPointer ; pointer to list of columns - level data
    defw TestLevel-1

.LowBlocks
; 1
    defb 192,128,128,128,128,128,128,128,128,128,128,128,128,192,128,128
    defb 192,0,0,0,0,0,0,0,0,0,0,0,0,192,0,0
    defb 192,0,0,0,0,0,0,0,0,0,0,0,0,192,0,0
    defb 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128
; 2
    defb 192,192,192,192,192,192,192,192,192,192,192,192,192,192,192,192
    defb 192,192,192,192,192,192,192,192,192,192,192,192,192,192,192,192
    defb 192,192,192,192,192,192,192,192,192,192,192,192,192,192,192,192
    defb 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128

.HighBlocks
;1
    defb 64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64
    defb 192,0,0,0,0,0,0,0,0,0,0,0,0,192,0,0
    defb 192,0,0,0,0,0,0,0,0,0,0,0,0,192,0,0
    defb 192,64,64,64,64,64,64,64,64,64,64,64,64,192,64,64
; 2
    defb 64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64
    defb 192,192,192,192,192,192,192,192,192,192,192,192,192,192,192,192
    defb 192,192,192,192,192,192,192,192,192,192,192,192,192,192,192,192
    defb 192,192,192,192,192,192,192,192,192,192,192,192,192,192,192,192
